/* $Id: rail.c 5827 2006-08-09 22:09:09Z richk $ */

#include "stdafx.h"
#include "openttd.h"
#include "bridge_map.h"
#include "rail.h"
#include "station_map.h"
#include "tunnel_map.h"
#include "variables.h"

/* XXX: Below 3 tables store duplicate data. Maybe remove some? */
/* Maps a trackdir to the bit that stores its status in the map arrays, in the
 * direction along with the trackdir */
const byte _signal_along_trackdir[] = {
	0x80, 0x80, 0x80, 0x20, 0x40, 0x10, 0, 0,
	0x40, 0x40, 0x40, 0x10, 0x80, 0x20
};

/* Maps a trackdir to the bit that stores its status in the map arrays, in the
 * direction against the trackdir */
const byte _signal_against_trackdir[] = {
	0x40, 0x40, 0x40, 0x10, 0x80, 0x20, 0, 0,
	0x80, 0x80, 0x80, 0x20, 0x40, 0x10
};

/* Maps a Track to the bits that store the status of the two signals that can
 * be present on the given track */
const byte _signal_on_track[] = {
	0xC0, 0xC0, 0xC0, 0x30, 0xC0, 0x30
};

/* Maps a diagonal direction to the all trackdirs that are connected to any
 * track entering in this direction (including those making 90 degree turns)
 */
const TrackdirBits _exitdir_reaches_trackdirs[] = {
	TRACKDIR_BIT_X_NE | TRACKDIR_BIT_LOWER_E | TRACKDIR_BIT_LEFT_N,  /* DIAGDIR_NE */
	TRACKDIR_BIT_Y_SE | TRACKDIR_BIT_LEFT_S  | TRACKDIR_BIT_UPPER_E, /* DIAGDIR_SE */
	TRACKDIR_BIT_X_SW | TRACKDIR_BIT_UPPER_W | TRACKDIR_BIT_RIGHT_S, /* DIAGDIR_SW */
	TRACKDIR_BIT_Y_NW | TRACKDIR_BIT_RIGHT_N | TRACKDIR_BIT_LOWER_W  /* DIAGDIR_NW */
};

const Trackdir _next_trackdir[] = {
	TRACKDIR_X_NE,  TRACKDIR_Y_SE,  TRACKDIR_LOWER_E, TRACKDIR_UPPER_E, TRACKDIR_RIGHT_S, TRACKDIR_LEFT_S, INVALID_TRACKDIR, INVALID_TRACKDIR,
	TRACKDIR_X_SW,  TRACKDIR_Y_NW,  TRACKDIR_LOWER_W, TRACKDIR_UPPER_W, TRACKDIR_RIGHT_N, TRACKDIR_LEFT_N
};

/* Maps a trackdir to all trackdirs that make 90 deg turns with it. */
const TrackdirBits _track_crosses_trackdirs[] = {
	TRACKDIR_BIT_Y_SE     | TRACKDIR_BIT_Y_NW,                                                   /* TRACK_X     */
	TRACKDIR_BIT_X_NE     | TRACKDIR_BIT_X_SW,                                                   /* TRACK_Y     */
	TRACKDIR_BIT_RIGHT_N  | TRACKDIR_BIT_RIGHT_S  | TRACKDIR_BIT_LEFT_N  | TRACKDIR_BIT_LEFT_S,  /* TRACK_UPPER */
	TRACKDIR_BIT_RIGHT_N  | TRACKDIR_BIT_RIGHT_S  | TRACKDIR_BIT_LEFT_N  | TRACKDIR_BIT_LEFT_S,  /* TRACK_LOWER */
	TRACKDIR_BIT_UPPER_W  | TRACKDIR_BIT_UPPER_E  | TRACKDIR_BIT_LOWER_W | TRACKDIR_BIT_LOWER_E, /* TRACK_LEFT  */
	TRACKDIR_BIT_UPPER_W  | TRACKDIR_BIT_UPPER_E  | TRACKDIR_BIT_LOWER_W | TRACKDIR_BIT_LOWER_E  /* TRACK_RIGHT */
};

/* Maps a track to all tracks that make 90 deg turns with it. */
const TrackBits _track_crosses_tracks[] = {
	TRACK_BIT_Y,    /* TRACK_X     */
	TRACK_BIT_X,    /* TRACK_Y     */
	TRACK_BIT_VERT, /* TRACK_UPPER */
	TRACK_BIT_VERT, /* TRACK_LOWER */
	TRACK_BIT_HORZ, /* TRACK_LEFT  */
	TRACK_BIT_HORZ  /* TRACK_RIGHT */
};

/* Maps a trackdir to the (4-way) direction the tile is exited when following
 * that trackdir */
const DiagDirection _trackdir_to_exitdir[] = {
	DIAGDIR_NE,DIAGDIR_SE,DIAGDIR_NE,DIAGDIR_SE,DIAGDIR_SW,DIAGDIR_SE, DIAGDIR_NE,DIAGDIR_NE,
	DIAGDIR_SW,DIAGDIR_NW,DIAGDIR_NW,DIAGDIR_SW,DIAGDIR_NW,DIAGDIR_NE,
};

const Trackdir _track_exitdir_to_trackdir[][DIAGDIR_END] = {
	{TRACKDIR_X_NE,     INVALID_TRACKDIR,  TRACKDIR_X_SW,     INVALID_TRACKDIR},
	{INVALID_TRACKDIR,  TRACKDIR_Y_SE,     INVALID_TRACKDIR,  TRACKDIR_Y_NW},
	{TRACKDIR_UPPER_E,  INVALID_TRACKDIR,  INVALID_TRACKDIR,  TRACKDIR_UPPER_W},
	{INVALID_TRACKDIR,  TRACKDIR_LOWER_E,  TRACKDIR_LOWER_W,  INVALID_TRACKDIR},
	{INVALID_TRACKDIR,  INVALID_TRACKDIR,  TRACKDIR_LEFT_S,   TRACKDIR_LEFT_N},
	{TRACKDIR_RIGHT_N,  TRACKDIR_RIGHT_S,  INVALID_TRACKDIR,  INVALID_TRACKDIR}
};

const Trackdir _track_enterdir_to_trackdir[][DIAGDIR_END] = { // TODO: replace magic with enums
	{TRACKDIR_X_NE,     INVALID_TRACKDIR,  TRACKDIR_X_SW,     INVALID_TRACKDIR},
	{INVALID_TRACKDIR,  TRACKDIR_Y_SE,     INVALID_TRACKDIR,  TRACKDIR_Y_NW},
	{INVALID_TRACKDIR,  TRACKDIR_UPPER_E,  TRACKDIR_UPPER_W,  INVALID_TRACKDIR},
	{TRACKDIR_LOWER_E,  INVALID_TRACKDIR,  INVALID_TRACKDIR,  TRACKDIR_LOWER_W},
	{TRACKDIR_LEFT_N,   TRACKDIR_LEFT_S,   INVALID_TRACKDIR,  INVALID_TRACKDIR},
	{INVALID_TRACKDIR,  INVALID_TRACKDIR,  TRACKDIR_RIGHT_S,  TRACKDIR_RIGHT_N}
};

const Trackdir _track_direction_to_trackdir[][DIR_END] = {
	{INVALID_TRACKDIR, TRACKDIR_X_NE,     INVALID_TRACKDIR, INVALID_TRACKDIR,  INVALID_TRACKDIR, TRACKDIR_X_SW,     INVALID_TRACKDIR, INVALID_TRACKDIR},
	{INVALID_TRACKDIR, INVALID_TRACKDIR,  INVALID_TRACKDIR, TRACKDIR_Y_SE,     INVALID_TRACKDIR, INVALID_TRACKDIR,  INVALID_TRACKDIR, TRACKDIR_Y_NW},
	{INVALID_TRACKDIR, INVALID_TRACKDIR,  TRACKDIR_UPPER_E, INVALID_TRACKDIR,  INVALID_TRACKDIR, INVALID_TRACKDIR,  TRACKDIR_UPPER_W, INVALID_TRACKDIR},
	{INVALID_TRACKDIR, INVALID_TRACKDIR,  TRACKDIR_LOWER_E, INVALID_TRACKDIR,  INVALID_TRACKDIR, INVALID_TRACKDIR,  TRACKDIR_LOWER_W, INVALID_TRACKDIR},
	{TRACKDIR_LEFT_N,  INVALID_TRACKDIR,  INVALID_TRACKDIR, INVALID_TRACKDIR,  TRACKDIR_LEFT_S,  INVALID_TRACKDIR,  INVALID_TRACKDIR, INVALID_TRACKDIR},
	{TRACKDIR_RIGHT_N, INVALID_TRACKDIR,  INVALID_TRACKDIR, INVALID_TRACKDIR,  TRACKDIR_RIGHT_S, INVALID_TRACKDIR,  INVALID_TRACKDIR, INVALID_TRACKDIR}
};

const Trackdir _dir_to_diag_trackdir[] = {
	TRACKDIR_X_NE, TRACKDIR_Y_SE, TRACKDIR_X_SW, TRACKDIR_Y_NW,
};


RailType GetTileRailType(TileIndex tile, Trackdir trackdir)
{
	DiagDirection exitdir = TrackdirToExitdir(trackdir);
	switch (GetTileType(tile)) {
		case MP_RAILWAY:
			return GetRailType(tile);

		case MP_STREET:
			/* rail/road crossing */
			if (IsLevelCrossing(tile)) return GetRailTypeCrossing(tile);
			break;

		case MP_STATION:
			if (IsRailwayStationTile(tile)) return GetRailType(tile);
			break;

		case MP_TUNNELBRIDGE:
			if (IsTunnel(tile)) {
				if (GetTunnelTransportType(tile) == TRANSPORT_RAIL) {
					return GetRailType(tile);
				}
			} else {
				if (IsBridgeRamp(tile)) {
					if (GetBridgeTransportType(tile) == TRANSPORT_RAIL) {
						return GetRailType(tile);
					}
				} else {
					if (GetBridgeAxis(tile) == DiagDirToAxis(exitdir)) {
						if (GetBridgeTransportType(tile) == TRANSPORT_RAIL) {
							return GetRailTypeOnBridge(tile);
						}
					} else {
						if (IsTransportUnderBridge(tile) &&
								GetTransportTypeUnderBridge(tile) == TRANSPORT_RAIL) {
							return GetRailType(tile);
						}
					}
				}
			}
			break;

		default:
			break;
	}
	return INVALID_RAILTYPE;
}

Trackdir GetNextTrackdir(TileIndex tile, Trackdir prev_dir)
{
	Trackdir dir;
	TrackBits track;
	DiagDirection entrydir = TrackdirToExitdir(prev_dir);

	if (GetTileType(tile) == MP_RAILWAY) {
		track = GetTrackBits(tile);
	} else if (GetTileType(tile) == MP_STREET && IsLevelCrossing(tile)) {
		track = GetCrossingRailBits(tile);
	} else {
		return INVALID_TRACKDIR;
	}

	// Find the Trackdir of this tile. The checks against entrydir work
	// out how we got on to this tile, so we can decide which track to take
	// from here.
	switch (track) {
		case TRACK_BIT_X:
			if (entrydir == DIAGDIR_NW || entrydir == DIAGDIR_SE)
				return INVALID_TRACKDIR;
			dir = (entrydir == DIAGDIR_NE) ? TRACKDIR_X_NE : TRACKDIR_X_SW;
			break;
		case TRACK_BIT_Y:
			if (entrydir == DIAGDIR_NE || entrydir == DIAGDIR_SW)
				return INVALID_TRACKDIR;
			dir = (entrydir == DIAGDIR_NW) ? TRACKDIR_Y_NW : TRACKDIR_Y_SE;
			break;
		case TRACK_BIT_HORZ:
		case TRACK_BIT_UPPER:
			if (entrydir == DIAGDIR_SE) {
				if (prev_dir == TRACKDIR_LEFT_S && _patches.forbid_90_deg)
					return INVALID_TRACKDIR;
				dir = TRACKDIR_UPPER_E;
				break;
			} else if (entrydir == DIAGDIR_SW) {
				if (prev_dir == TRACKDIR_RIGHT_S && _patches.forbid_90_deg)
					return INVALID_TRACKDIR;
				dir = TRACKDIR_UPPER_W;
				break;
			} else if (track == TRACK_BIT_UPPER) {
				return INVALID_TRACKDIR;
			}
			/* FALLTHROUGH */
		case TRACK_BIT_LOWER:
			if (entrydir == DIAGDIR_NE) {
				if (prev_dir == TRACKDIR_RIGHT_N && _patches.forbid_90_deg)
					return INVALID_TRACKDIR;
				dir = TRACKDIR_LOWER_E;
			} else if (entrydir == DIAGDIR_NW) {
				if (prev_dir == TRACKDIR_LEFT_N && _patches.forbid_90_deg)
					return INVALID_TRACKDIR;
				dir = TRACKDIR_LOWER_W;
			} else {
				return INVALID_TRACKDIR;
			}
			break;
		case TRACK_BIT_VERT:
		case TRACK_BIT_LEFT:
			if (entrydir == DIAGDIR_NE) {
				if (prev_dir == TRACKDIR_UPPER_E && _patches.forbid_90_deg)
					return INVALID_TRACKDIR;
				dir = TRACKDIR_LEFT_N;
				break;
			} else if (entrydir == DIAGDIR_SE) {
				if (prev_dir == TRACKDIR_LOWER_E && _patches.forbid_90_deg)
					return INVALID_TRACKDIR;
				dir = TRACKDIR_LEFT_S;
				break;
			} else if (track == TRACK_BIT_LEFT) {
				return INVALID_TRACKDIR;
			}
			/* FALLTHROUGH */
		case TRACK_BIT_RIGHT:
			if (entrydir == DIAGDIR_NW) {
				if (prev_dir == TRACKDIR_UPPER_W && _patches.forbid_90_deg)
					return INVALID_TRACKDIR;
				dir = TRACKDIR_RIGHT_N;
				break;
			} else if (entrydir == DIAGDIR_SW) {
				if (prev_dir == TRACKDIR_LOWER_W && _patches.forbid_90_deg)
					return INVALID_TRACKDIR;
				dir = TRACKDIR_RIGHT_S;
				break;
			} else {
				return INVALID_TRACKDIR;
			}
		default:
			return INVALID_TRACKDIR;
	}
	return dir;
}

TileIndex RailLookAhead(TileIndex tile, Trackdir dir, uint count)
{
	uint i;
	DiagDirection next_tile;

	for (i = 0; i < count; i++) {
		next_tile = TrackdirToExitdir(dir);
		tile += TileOffsByDir(next_tile);
		dir = GetNextTrackdir(tile, dir);
		if (dir == INVALID_TRACKDIR) return 0;
	}

	return tile;
}
